from typing import Optional

from playhouse.sqlite_ext import FTSModel, IntegerField, SearchField

import sanguine.constants as c
from sanguine.db import db
from sanguine.state import get_counter, update_counter

type_to_id = {c.ENTITY_VARIABLE: 1, c.ENTITY_FUNCTION: 2, c.ENTITY_CLASS: 3}
id_to_type = {v: k for k, v in type_to_id.items()}


class CodeEntity(FTSModel):
    id = IntegerField()
    type = IntegerField()
    file = SearchField()
    name = SearchField()

    class Meta:
        database = db


def fts_add_symbol(path: str, type: int, name: str) -> Optional[CodeEntity]:
    exists = (
        CodeEntity.select()
        .where(
            (CodeEntity.file == path)
            & (CodeEntity.name == name)
            & CodeEntity.type
            == type
        )
        .exists()
    )
    if exists:
        return

    update_counter()
    entity = CodeEntity.create(
        id=get_counter(),
        file=path,
        type=type,
        name=name,
    )
    return entity


def fts_remove_symbol(path: str, type: int, name: str) -> list[int]:
    query = CodeEntity.select(CodeEntity.id).where(
        (CodeEntity.file == path)
        & (CodeEntity.type == type)
        & (CodeEntity.name == name)
    )
    ids = [obj.id for obj in query]
    if ids:
        CodeEntity.delete().where(CodeEntity.id.in_(ids)).execute()
    return ids



import argparse
import sys

import sanguine.constants as c
import sanguine.meta as meta


def main():
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        "--cuda", action="store_true", help="Use GPU to run embedding model"
    )
    global_parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=f"%(prog)s {meta.version}",
        help="Show version and exit",
    )

    parser = argparse.ArgumentParser(
        prog=meta.name,
        description=f"Keep it D.R.Y with {meta.name}.",
        formatter_class=lambda prog: argparse.HelpFormatter(
            prog, max_help_position=25, width=120
        ),
        usage=argparse.SUPPRESS,
        parents=[global_parser],
    )

    subparsers = parser.add_subparsers(
        dest="command", help="Available commands"
    )

    subparsers.add_parser(
        "install", help="Install post-commit hook", parents=[global_parser]
    )
    subparsers.add_parser(
        "uninstall", help="Uninstall post-commit hook", parents=[global_parser]
    )

    index_parser = subparsers.add_parser(
        "index", help="Index code", parents=[global_parser]
    )
    index_parser.add_argument(
        "--commit-id",
        "-c",
        type=str,
        help="Specific commit ID to index (defaults to last commit)",
    )
    index_parser.add_argument(
        "--file",
        "-f",
        type=str,
        help="Specific file to index",
    )
    index_parser.add_argument(
        "--all-files",
        "-a",
        action="store_true",
        help="Index all files in current folder not ignored by .gitignore",
    )

    search_parser = subparsers.add_parser(
        "search", help="Search indexed code", parents=[global_parser]
    )
    search_parser.add_argument(
        "text", type=str, nargs="?", help="Search query"
    )
    search_parser.add_argument(
        "--count", "-k", type=int, default=10, help="Number of results"
    )
    search_parser.add_argument(
        "--interactive",
        "-i",
        action="store_true",
        help="Interactive (REPL) search",
    )
    search_parser.add_argument(
        "--path", "-p", type=str, help="Filter by path prefix"
    )
    search_parser.add_argument(
        "--type",
        "-t",
        type=str,
        choices=[c.ENTITY_CLASS, c.ENTITY_FUNCTION],
        help="Filter by code entity type",
    )

    delete_parser = subparsers.add_parser(
        "delete", help="Delete indexed entities", parents=[global_parser]
    )
    delete_parser.add_argument(
        "--name",
        "-n",
        type=str,
        help="Name or partial name of entities to delete",
    )
    delete_parser.add_argument(
        "--path", "-p", type=str, help="Path prefix to filter entities"
    )
    delete_parser.add_argument(
        "--type",
        "-t",
        type=str,
        choices=[c.ENTITY_CLASS, c.ENTITY_FUNCTION],
        help="Filter by entity type",
    )
    delete_parser.add_argument(
        "--yes", "-y", type=str, help="Delete without asking"
    )

    subparsers.add_parser(
        "refresh", help="Refresh the embedding index", parents=[global_parser]
    )

    args = parser.parse_args(sys.argv[1:] or ["-h"])

    # --- imports here, so '--help' doesn't do any heavy-lifting ---
    from colorama import Fore, Style
    from colorama import init as colorama_init

    from sanguine.db import db
    from sanguine.db.fts import CodeEntity
    from sanguine.install_uninstall import install, uninstall

    colorama_init(autoreset=True)

    if db.is_closed():
        db.connect()
    db.create_tables([CodeEntity], safe=True)

    from sanguine.core import (
        delete,
        index_all_files,
        index_file,
        process_commit,
        search,
    )
    from sanguine.db.hnsw import init_embedder, refresh_hnsw_index, save_indices

    init_embedder(use_cuda=args.cuda)

    if args.command == "install":
        install()

    elif args.command == "uninstall":
        uninstall()

    elif args.command == "index":
        if args.file:
            index_file(args.file)
        elif args.all_files:
            index_all_files()
        else:
            process_commit(args.commit_id)

        save_indices()

    elif args.command == "search":
        if not args.interactive:
            if not args.text:
                print("Search text is required unless using --interactive")
                sys.exit(1)
            search(args.text, k=args.count, path=args.path, type=args.type)
            return
        print("Interactive search. Type ':q' to quit.")
        while True:
            cmd = input(">> ").strip()
            if cmd.lower() == ":q":
                break
            parts = cmd.split()
            if not parts:
                continue
            kwargs = search_parser.parse_args(parts)
            search(
                kwargs.text, k=kwargs.count, path=kwargs.path, type=kwargs.type
            )

    elif args.command == "delete":
        if not args.name and not args.path:
            print(
                f"{Fore.RED}Error: You must provide at least --name or --path for deletion.{Style.RESET_ALL}"
            )
            sys.exit(1)
        delete(name=args.name, path=args.path, type=args.type, force=args.yes)
        save_indices()

    elif args.command == "refresh":
        refresh_hnsw_index()

    else:
        parser.print_help()
        sys.exit(1)



import os
import shutil
import stat
import textwrap

from colorama import Fore, Style
from colorama import init as colorama_init

import sanguine.meta as meta
from sanguine.utils import is_repo

colorama_init(autoreset=True)


def install():
    if not is_repo():
        print(f"{Fore.RED}Not a git repository.{Style.RESET_ALL}")
        return

    hooks_dir = ".git/hooks"
    hook_file = os.path.join(hooks_dir, "post-commit")
    os.makedirs(hooks_dir, exist_ok=True)

    py_exe = shutil.which("python")
    if py_exe:
        py_exe = py_exe.replace("\\", "/")

    script = textwrap.dedent(
        f"""\
        #!/bin/sh
        export FORCE_COLOR=1
        export TERM=xterm-256color

        PYTHON="{py_exe}"
        PREFIX=""

        if [ -x "$PYTHON" ]; then
            exec "$PYTHON" -m {meta.name} index
        elif command -v {meta.name} > /dev/null; then
            exec {meta.name} index
        else
            echo "{meta.name} not found"
            exit 1
        fi
        """
    )

    with open(hook_file, "w") as f:
        f.write(script)

    os.chmod(hook_file, os.stat(hook_file).st_mode | stat.S_IXUSR)
    print(f"{Fore.GREEN}{meta.name} has been installed!{Style.RESET_ALL}")


def uninstall():
    hook_file = ".git/hooks/post-commit"
    if os.path.exists(hook_file):
        os.remove(hook_file)
        print(
            f"{Fore.GREEN}{meta.name} has been uninstalled!{Style.RESET_ALL}"
        )
    else:
        print(f"{Fore.YELLOW}No hook found to uninstall.{Style.RESET_ALL}")



import os

from peewee import SqliteDatabase

import sanguine.meta as meta
from sanguine.utils import app_dir

os.makedirs(app_dir, exist_ok=True)
db_file = os.path.join(app_dir, "db.db")
db = SqliteDatabase(db_file)



import subprocess


def run_command(command: list[str]) -> str:
    return subprocess.run(
        command, capture_output=True, text=True, check=True
    ).stdout.rstrip("\n")


def last_commit() -> str:
    return run_command(["git", "rev-parse", "HEAD"])


def changed_files(commit_id: str) -> list[str]:
    result = run_command(
        [
            "git",
            "diff-tree",
            "--root",
            "--no-commit-id",
            "--name-only",
            "-r",
            commit_id,
        ]
    )
    return [f.strip() for f in result.split("\n") if f.strip()]


def get_commit_file_changes(commit_id: str, file: str) -> str:
    return run_command(["git", "show", commit_id, "--", file])


def commit_diff(commit_id: str) -> dict[str, tuple[str, str]]:
    files = changed_files(commit_id)
    diff = [get_commit_file_changes(commit_id, file) for file in files]
    file_diff = {}

    for file, diff in zip(files, diff):
        diff_lines = diff.split("\n")
        added_lines = []
        removed_lines = []

        for line in diff_lines:
            if line.startswith("+") and not line.startswith("+++"):
                added_lines.append(line[1:])
            elif line.startswith("-") and not line.startswith("---"):
                removed_lines.append(line[1:])
        file_diff[file] = ("\n".join(added_lines), "\n".join(removed_lines))

    return file_diff



import difflib
import os
import subprocess
import sys
from functools import reduce
from typing import Optional

import pathspec
from colorama import Fore, Style
from tqdm import tqdm

import sanguine.constants as c
import sanguine.git as git
import sanguine.meta as meta
from sanguine.db import db
from sanguine.db.fts import (
    CodeEntity,
    fts_add_symbol,
    fts_remove_symbol,
    id_to_type,
    type_to_id,
)
from sanguine.db.hnsw import indices as hnsw_indices, hnsw_add_symbol, hnsw_remove_symbol, hnsw_search, find_apt_hnsw_index, make_hnsw_index
from sanguine.parser import extract_symbols
from sanguine.state import get_staleness, update_staleness
from sanguine.utils import ext_to_lang, is_repo, normalize_path


def index_diff(file_diff: dict[str, tuple[str, Optional[str]]], dir: str):
    dir = normalize_path(dir)
    hnsw_index = find_apt_hnsw_index(dir)
    hnsw_index = hnsw_index or make_hnsw_index(dir)

    for file, (added_lines, removed_lines) in tqdm(
        file_diff.items(),
        total=len(file_diff),
        ncols=80,
        bar_format=f"{Fore.GREEN}[{meta.name}] |{{bar}}|{Style.RESET_ALL}",
    ):
        ext = os.path.splitext(file)[1]
        if ext not in ext_to_lang:
            continue

        file_path = normalize_path(file)
        lang = ext_to_lang[ext]

        added_symbols = extract_symbols(added_lines, lang)
        removed_symbols = {c.FLD_FUNCTIONS: [], c.FLD_CLASSES: []}
        if removed_lines:
            removed_symbols = extract_symbols(removed_lines, lang)

        with db.atomic():
            for entity_type, field_name in [
                (c.ENTITY_FUNCTION, c.FLD_FUNCTIONS),
                (c.ENTITY_CLASS, c.FLD_CLASSES),
            ]:
                for symbol_name in added_symbols[field_name]:
                    o = fts_add_symbol(
                        path=file_path,
                        type=type_to_id[entity_type],
                        name=symbol_name,
                    )
                    hnsw_add_symbol([symbol_name], [o.id], hnsw_index)

                for symbol_name in removed_symbols[field_name]:
                    ids = fts_remove_symbol(
                        file_path, type_to_id[entity_type], symbol_name
                    )
                    for id in ids:
                        hnsw_remove_symbol(id)
    print()


def process_commit(commit_id: Optional[str] = None):
    if not is_repo():
        print(
            f"{Fore.RED}Error: not a git repository.{Style.RESET_ALL}",
            file=sys.stderr,
        )
        return

    try:
        commit = commit_id or git.last_commit()
        file_to_diff = git.commit_diff(commit)
        index_diff(file_to_diff, os.getcwd())
    except subprocess.CalledProcessError:
        print(
            f"{Fore.RED}Invalid commit ID{Style.RESET_ALL}",
            file=sys.stderr,
        )


def index_file(file: str):
    if not os.path.isfile(file):
        print(
            f"{Fore.RED}{file} is not a file{Style.RESET_ALL}",
            file=sys.stderr,
        )
        return
    with open(file, encoding="utf-8") as f:
        index_diff({file: (f.read(), "")}, os.path.dirname(file))


def index_all_files():
    cwd = os.getcwd()
    ignore_file = os.path.join(cwd, ".gitignore")
    patterns = []
    if os.path.exists(ignore_file):
        with open(ignore_file, "r") as f:
            patterns = f.read().splitlines()

    spec = pathspec.PathSpec.from_lines("gitwildmatch", patterns)
    file_diff = {}

    for dirpath, dirnames, filenames in os.walk(cwd):
        if ".git" in dirnames:
            dirnames.remove(".git")
        dirnames[:] = [d for d in dirnames if not d.startswith(".")]

        for name in filenames:
            if name.startswith("."):
                continue

            filepath = os.path.relpath(os.path.join(dirpath, name), cwd)
            if spec.match_file(filepath):
                continue

            try:
                with open(filepath) as f:
                    file_diff[filepath] = (f.read(), "")
            except (UnicodeDecodeError, PermissionError):
                continue

    if not file_diff:
        print(f"{Fore.YELLOW}No indexable files found.{Style.RESET_ALL}")
        return

    index_diff(file_diff, cwd)


def search(
    query: str,
    k: int,
    path: Optional[str] = None,
    type: Optional[str] = None,
    show_score: bool = True,
):
    conditions = [CodeEntity.name.contains(query)]
    if path is not None:
        path = normalize_path(path)
        conditions.append(CodeEntity.file.startswith(path))
    if type is not None:
        type = type_to_id[type]
        conditions.append(CodeEntity.type == type)

    conditions = reduce(lambda x, y: x & y, conditions)
    db_objects = CodeEntity.select().where(conditions)
    db_map = {o.id: o for o in db_objects}

    total_hnsw_no, stale_hnsw_no = 0, 0
    sim_ids, sim_scores = [], []

    if path is not None:
        # use the parent directory's index
        matching_index = find_apt_hnsw_index(path)
        if matching_index is not None:
            sim_ids, sim_scores = hnsw_search(query, matching_index, k=k)
    else:
        # use all indices
        for _, index in hnsw_indices.items():
            _sim_ids, _sim_scores = hnsw_search(query, index, k=k)
            sim_ids.extend(_sim_ids), sim_scores.extend(_sim_scores)
            scored_ids = sorted(zip(sim_ids, sim_scores), key=lambda x: x[1], reverse=True)[:k]
            sim_ids, sim_scores = [list(x) for x in zip(*scored_ids)]


    total_hnsw_no += len(sim_ids)
    sim_score_map = dict(zip(sim_ids, sim_scores))
    sim_map = {
        o.id: o for o in CodeEntity.select().where(CodeEntity.id.in_(sim_ids))
    }
    stale_hnsw_no += len(sim_ids) - len(sim_map)
    sim_ids = set(sim_ids)

    if len(sim_map) < k / 2:
        more_ids, more_scores = hnsw_search(query, k=k * 2)
        for s_id, score in zip(more_ids, more_scores):
            if s_id in sim_ids:
                continue
            total_hnsw_no += 1
            s_o = CodeEntity.get_or_none(id=s_id)
            if s_o is None:
                stale_hnsw_no += 1
                continue
            sim_ids.add(s_id)
            sim_score_map[s_id] = score
            sim_map[s_id] = s_o

    if total_hnsw_no > 0:
        update_staleness(total_hnsw_no, stale_hnsw_no)
        staleness = get_staleness()

    if staleness > 0.5:
        print(
            f"{Fore.YELLOW}HNSW needs index refreshing, >50% entries are stale. Bordering uselessness.{Style.RESET_ALL}"
        )
    elif staleness > 0.3:
        print(
            f"{Fore.YELLOW}HNSW needs rindex efreshing, >30% entries are stale.{Style.RESET_ALL}"
        )
    if staleness > 0.3:
        print(f'run "{meta.name} refresh" to refresh\n')

    all_ids = list(sim_ids) + [o.id for o in db_objects if o.id not in sim_ids]
    results = []
    for oid in all_ids:
        obj = sim_map.get(oid) or db_map.get(oid)
        if obj is None or (type is not None and obj.type != type):
            continue
        sim_score = sim_score_map.get(oid, 0)
        text_score = difflib.SequenceMatcher(
            None, query.lower(), obj.name.lower()
        ).ratio()
        final_score = sim_score + text_score
        results.append((obj, final_score))

    results.sort(key=lambda x: x[1], reverse=True)
    results = results[:k]

    if not results:
        print("No matches found.")
        return

    last_file = None
    for obj, score in results:
        if obj.file != last_file:
            filename = f"{Fore.CYAN}{obj.file}{Style.RESET_ALL}"
            if last_file is not None:
                filename = "\n" + filename
            print(filename)

        color = (
            Fore.GREEN
            if id_to_type[obj.type] == c.ENTITY_FUNCTION
            else Fore.BLUE
        )
        line = f"  {color}↳ {obj.name}{Style.RESET_ALL}"
        if show_score:
            line += f" ({score:.2f})"
        print(line)
        last_file = obj.file

    print()


def delete(
    name: Optional[str] = None,
    path: Optional[str] = None,
    type: Optional[str] = None,
    force: bool = False,
):
    conditions = []
    if name:
        conditions.append(CodeEntity.name.startswith(name))
    if path:
        path = normalize_path(path)
        conditions.append(CodeEntity.file.startswith(path))
    if type:
        conditions.append(CodeEntity.type == type_to_id[type])

    if not conditions:
        print(
            f"{Fore.YELLOW}No criteria provided for deletion.{Style.RESET_ALL}"
        )
        return

    query = CodeEntity.select().where(reduce(lambda x, y: x & y, conditions))
    total_no = query.count()

    if total_no == 0:
        print("No matching entities found for deletion.")
        return

    if force:
        with db.atomic():
            for obj in query:
                deleted_ids = fts_remove_symbol(obj.file, obj.type, obj.name)
                for _id in deleted_ids:
                    hnsw_remove_symbol(_id)
        print(f"{Fore.GREEN}{total_no} entities deleted.{Style.RESET_ALL}")
        return

    while True:
        print(
            f"{Fore.YELLOW}Warning: {total_no} entities match the criteria!{Style.RESET_ALL}"
        )
        choice = (
            input("Type 'yes' or 'y' to delete, 'list' to preview entities: ")
            .strip()
            .lower()
        )

        if choice in {"yes", "y"}:
            with db.atomic():
                for obj in query:
                    deleted_ids = fts_remove_symbol(
                        obj.file, obj.type, obj.name
                    )
                    for _id in deleted_ids:
                        hnsw_remove_symbol(_id)
            print(f"{Fore.GREEN}{total_no} entities deleted.{Style.RESET_ALL}")
            break

        elif choice == "list":
            print(
                f"{Fore.CYAN}Listing all matching entities:{Style.RESET_ALL}"
            )
            for obj in query:
                color = (
                    Fore.GREEN
                    if id_to_type[obj.type] == c.ENTITY_FUNCTION
                    else Fore.BLUE
                )
                print(f"{color}↳ {obj.name}{Style.RESET_ALL} in {obj.file}")
            print()

        else:
            print("\nDeletion cancelled.")
            break



import base64
import json
import os

import appdirs

import sanguine.meta as meta

app_dir = appdirs.user_data_dir(meta.name)
core_dir = os.path.dirname(__file__)

prog_lang_schema = json.load(
    open(os.path.join(core_dir, "assets", "prog_langs_schema.json"))
)
ext_to_lang = json.load(
    open(os.path.join(core_dir, "assets", "ext_to_lang.json"))
)


def is_repo():
    return os.path.exists(".git")


def normalize_path(path: str) -> str:
    path = os.path.abspath(os.path.expanduser(path))
    return path.replace(os.sep, "/").rstrip("/")


def encode_path(path: str) -> str:
    return base64.urlsafe_b64encode(path.encode()).decode()


def decode_path(name: str) -> str:
    return base64.urlsafe_b64decode(name.encode()).decode()



import json
import os

from sanguine.utils import app_dir

counter_file = os.path.join(app_dir, "state.json")

COUNTER = "counter"
HNSW_STALE_ENCOUNTERS = "hnsw_stale_encounters"
HNSW_TOTAL_ENCOUNTERS = "hnsw_total_encounters"

state = {COUNTER: 0, HNSW_STALE_ENCOUNTERS: 0, HNSW_TOTAL_ENCOUNTERS: 0}

if not os.path.isfile(counter_file):
    with open(counter_file, "w") as f:
        json.dump(state, f)
else:
    with open(counter_file) as f:
        state = json.load(f)


def save():
    with open(counter_file, "w") as f:
        json.dump(state, f)


def update_counter():
    state[COUNTER] += 1
    save()


def get_counter() -> int:
    return state[COUNTER]


def update_staleness(total: int, stale: int):
    state[HNSW_TOTAL_ENCOUNTERS] += total
    state[HNSW_STALE_ENCOUNTERS] += stale
    save()


def reset_staleness_metrics():
    state[HNSW_TOTAL_ENCOUNTERS] = 0
    state[HNSW_STALE_ENCOUNTERS] = 0
    save()


def get_staleness() -> float:
    return state[HNSW_STALE_ENCOUNTERS] / state[HNSW_TOTAL_ENCOUNTERS]



import os
from typing import Union

import hnswlib
import numpy as np
from colorama import Fore, Style
from fastembed import TextEmbedding
from tqdm import tqdm

from sanguine.db.fts import CodeEntity
from sanguine.utils import app_dir, normalize_path, encode_path, decode_path
from typing import Optional

dim = 384
model: Union[None, TextEmbedding] = None
indices_dir = os.path.join(app_dir, "indices")
indices = {}


if os.path.isdir(indices_dir):
    for index_file in os.listdir(indices_dir):
        index = hnswlib.Index(space="cosine", dim=dim)
        index_file = os.path.join(indices_dir, index_file)
        encoded_repo_path = os.path.splitext(os.path.basename(index_file))[0]
        index.load_index(index_file)
        norm_repo_path = decode_path(encoded_repo_path)
        indices[norm_repo_path] = index
else:
    os.makedirs(indices_dir)


def init_embedder(use_cuda: bool):
    global model

    providers = ["CPUExecutionProvider"]
    if use_cuda:
        providers.append("CUDAExecutionProvider")

    model = TextEmbedding(
        model_name="sentence-transformers/all-MiniLM-L6-v2",
        providers=providers,
    )


def make_hnsw_index(repo_path: str) -> hnswlib.Index:
    num_entities = CodeEntity.select().where(CodeEntity.file.startswith(repo_path)).count()
    max_elements = max(1000, num_entities + 100)
    index = hnswlib.Index(space="cosine", dim=dim)
    index.init_index(max_elements=max_elements, M=64)
    indices[repo_path] = index
    return index


def find_apt_hnsw_index(path: str) -> Optional[hnswlib.Index]:
    for repo_path, index in indices.items():
        # NOTE: 'normalize_path' here is for handling trailing slashes
        if normalize_path(os.path.commonpath([repo_path, path])) == repo_path:
            return index
    return None


def embed(texts: str) -> list[np.ndarray]:
    return list(model.embed(texts))


def hnsw_add_symbol(texts: list[str], ids: list[int], index: hnswlib.Index):
    embeddings = embed(texts)
    new_count = index.get_current_count() + len(ids)
    if new_count > index.get_max_elements():
        index.resize_index(max(new_count, index.get_max_elements() * 2))
    index.add_items(embeddings, ids)


def hnsw_search(query: str, index: hnswlib.Index, k: int = 10) -> tuple[list[int], list[float]]:
    if index.get_current_count() == 0:
        return []

    index.set_ef(max(50, k * 2))
    query_vec = embed([query])
    labels, distances = index.knn_query(query_vec, k=k)
    return labels[0].tolist(), [1 - d for d in distances[0].tolist()]


def hnsw_remove_symbol(id: int, index: Optional[hnswlib.Index] = None):
    if index is None:
        return
    index.mark_deleted(id)


def refresh_hnsw_index(batch_size: int = 512):
    total_entities = CodeEntity.select().count()
    if total_entities == 0:
        return

    pbar = tqdm(
        total=total_entities,
        ncols=80,
        bar_format=f"{Fore.GREEN}|{{bar}}|{Style.RESET_ALL}",
    )
    for repo_path in indices:
        index = hnswlib.Index(space="cosine", dim=dim)
        index.init_index(max_elements=max(total_entities, 1000), M=64)
        batch_ids, batch_texts = [], []

        for entity in CodeEntity.select().where(CodeEntity.file.startswith(repo_path)).iterator():
            batch_ids.append(entity.id)
            batch_texts.append(entity.name)

            if len(batch_ids) >= batch_size:
                embeddings = embed(batch_texts)
                index.add_items(embeddings, batch_ids)
                batch_ids.clear()
                batch_texts.clear()
                pbar.update(len(batch_ids))

        if batch_ids:
            embeddings = embed(batch_texts)
            index.add_items(embeddings, batch_ids)
            pbar.update(len(batch_ids))

        indices[repo_path] = index

    save_indices()
    print()


def save_indices():
    for repo_path, index in indices.items():
        filename = f"{encode_path(repo_path)}.bin"
        index.save_index(os.path.join(indices_dir, filename))



from tree_sitter_language_pack import get_parser

import sanguine.constants as c
from sanguine.utils import prog_lang_schema


def extract_symbols(code: str, lang: str) -> dict:
    result = {c.FLD_FUNCTIONS: [], c.FLD_CLASSES: []}
    nodes = prog_lang_schema.get(lang.lower())

    if not nodes:
        return result

    parser = get_parser(lang)
    code_bytes = code.encode("utf8")
    tree = parser.parse(code_bytes)
    root = tree.root_node

    def extract_name(node):
        for child in node.children:
            if child.type == nodes["identifier"]:
                return code_bytes[child.start_byte : child.end_byte].decode(
                    "utf8"
                )
        return None

    def traverse(node):
        if node.type == nodes["function"]:
            name = extract_name(node)
            if name:
                result[c.FLD_FUNCTIONS].append(name)

        elif node.type == nodes["class"]:
            name = extract_name(node)
            if name:
                result[c.FLD_CLASSES].append(name)

        for child in node.children:
            traverse(child)

    traverse(root)
    return result
